new;

/* ******************************************************************
*  This program provides estimates presented in KMSS (2017). 
*  The program performs the following estimators 
*
*  1. 2 WAY FIXED EFFECT: country pair and time FE
*  2. CTFE: country time fixed effect within estimator proposed in BMW(2017) and KMSS(2017) eq (10.5)  
*  3. 2D CCEP: 2D CCEP as proposed in Pesaran (2006) and KMSS(2017) eq (10.11)  
*  4. 3D CCEP: 3D CCEP as proposed in  KMSS(2017) eq (10.18) 
*
*  This program also performs the CSD test as presented in KMSS(2017) eq (10.34) and the alpha exponent test as presented in BKP (2015):
*  This program consists of the following steps:

	STEP 1:  loading data and necessary procedure ".src" files
	STEP 2: selecting the relevant dataset 
    STEP 3: data handling for within trasformations, group means
	STEP 3: estimation and CSD test of the 4 above mentioned estimators

*  Results are saved in excel files
*			
* ************************************************************************************** */
cls;

// ******************************************************
// STEP 1:  loading data and necessary procedure ".src" files
// ******************************************************

#include code_ord_holm.src;

n = 14;
t = 49;
cold= 15;


data = xlsreadm("datakmss","A2:O8919",1,""); 

// ******************************************************
// STEP 2: selecting the relevant dataset 
// ******************************************************

@ selecting index i home country; index j destination country; index t for years @

ind = data[.,1:3];          

@ selecting dependent variable @
y = data[.,4];

@ selecting regressors @
x = data[.,5:9]~data[.,11]~data[.,10];

Xname = "gdph"|"gdpf"|"sim"|"rlf"|"rer"|"cee"|"euro";
Xname1 = "sim"|"rlf"|"rer"|"cee"|"euro";
Xname2 = "gdph"|"gdpf"|"sim"|"rlf"|"rer"|"cee"|"euro"|"const";



@ definition of relevant parameters @

k = cols(x);
kk = k -1;
st = ind[.,3]-1959;
ni = n*(n-1)*t;

nj = 14;
ni = 13; 

nit = (n-1)*t;
njt = n*t;
nn = n*(n-1);


iy =  ind~y;

ind3 = seqa(1,1,rows(y));

ind2ti =  ind[.,3].*100+ind[.,1];                @ index for t & i @
ind2tj =  ind[.,3].*100+ind[.,2];                @ index for t & j @
ind2ij =  ind[.,1].*100+ind[.,2];                @ index for i & j @

i2yij =  ind3~ind2ij~y;   i2yj  =  ind3~ind~y;   i2yit =  ind3~ind2ti~y;  i2yjt =  ind3~ind2tj~y;

// ******************************************************
//    STEP 3: data handling for within trasformations, group means
// ******************************************************

myi   = TRDJT(y, nn, t);            @ y_bar_i @
myj   = TRDJ(i2yj, nj, nit);       @ SORT y_bar_j @
myt   = TRDT(i2yj, t, nn);         @ SORT y_bar_t @
myij  = TRDIJ(i2yij, nn, t);              @ SORT y_bar_ij @
myit  = TRDITS(i2yit, njt, ni);         @ SORT y_bar_it  @
myjt  = TRDJTS(i2yjt, njt, ni);         @ SORT y_bar_jt  @
gmy = meanc(y);

gmx = zeros(rows(y),k);
j=1; do while j<=k;
    gmx[.,j] = meanc(x[.,j]).*ones(rows(y),1);
j=j+1;endo;
mxij = zeros(rows(y),k);
j=1; do while j<=k;
    mxij[.,j] = TRDIJ(ind3~ind2ij~x[.,j], nn, t);
j=j+1;endo;
mxjt = zeros(rows(y),k);
j=1; do while j<=k;
    mxjt[.,j] = TRDJTS(ind3~ind2tj~x[.,j], njt, ni);
j=j+1;endo;
mxit = zeros(rows(y),k);
j=1; do while j<=k;
    mxit[.,j] = TRDITS(ind3~ind2ti~x[.,j],  njt, ni);
j=j+1;endo;
mxt = zeros(rows(y),k);
j=1; do while j<=k;
    mxt[.,j] = TRDT(ind3~ind~x[.,j], t, nn);
j=j+1;endo;
mxi = zeros(rows(y),k);
j=1; do while j<=k;
    mxi[.,j] = TRDJT(x[.,j], nn, t);
j=j+1;endo;
mxj = zeros(rows(y),k);
j=1; do while j<=k;
    mxj[.,j] = TRDJ(ind3~ind~x[.,j], nj, nit);
j=j+1;endo;

@  ************* 2 WAY  FIXED EFFECT ***********@

dyf2 = y - myij - myt + gmy;

dxf2 = zeros(rows(y),k);
j=1; do while j<=k;
    dxf2[.,j] = x[.,j] - mxij[.,j]  - mxt[.,j]  + gmx[.,j];
j=j+1;endo;

dxf2 =  dxf2[.,1:k];

{bfe2w,Qse,Qt, Qsig2} = FE2w(dyf2,dxf2,nn,T,k);

PROUT = Xname~bfe2w~Qse~Qt;
"BETA 2 WAY FE ";"             Coeff     se      t-ratio";
call printfm(PRout,0~1~1~1,"s"~8~8|"lf"~10~4|"lf"~10~4|"lf"~10~4); 
"";

@ output in excel 2 WAY POOL FE  @
ret = xlsWritem(PROUT, "2FEP", "b2", 1, 0); 


@ computing residuals @
ei = y - x*bfe2w;      
em = reshape(ei,nn,t);
ei = em';

@ computing CSD test and alpha exponent @

{t_res,n_res,cdx} = CDP(ei);
"CD test";cdx;

x_res = ei ; /*  T by N */
a_size=0.1;
x_resst=standx(x_res);

/********** Exponent of cross-sectional dependence ************/
{rej,a_res,v_f_res,rej,rej,ggg_res,s_res,rej}=atildeall(x_resst,a_size);
om_res=1.65*((1/t_res)*(v_f_res)+(4/n_res)*(n_res^(1-a_res)*s_res/(ggg_res-1)))^(1/2)/(2*ln(n_res)); /* 90% CI: 1.65, 95% CI: 1.96 */


"";"";
" Exponents of CSD (alpha) and c.i. 90% ";
"a_res"; (a_res-om_res)~a_res~(a_res+om_res);
"";"";
/*
@ excel output for CSD test@
ret = xlsWritem(cdx, "2FEP", "c10", 1, 0); 
ret = xlsWritem((a_res-om_res)~a_res~(a_res+om_res), "2FEP", "c12", 1, 0); 
*/

@  ************* 2D PCCE ***********@

@ factors selected for the  final version@

ff = mxt[.,2]~mxt[.,3:4]~mxt[.,6]~st~data[.,15]; 

dyf  = trp(y); 

dxf = zeros(rows(x),k);
j=1; do while j<=k;
  dxf[.,j] = trp(x[.,j]);
j=j+1;endo;

dxf =  dxf;

{bpcce, sehet, tstat} = FE1(dyf,dxf,nn,T,k);

bccep = bpcce~sehet~tstat;
PROUT = Xname~bccep ;
"BETA 2D PCCE ";"             Coeff     se      t-ratio";
call printfm(PRout,0~1~1~1,"s"~8~8|"lf"~10~4|"lf"~10~4|"lf"~10~4); 
"";

@ output in excel   @
ret = xlsWritem(PROUT, "2DPCCE", "b2", 1, 0); 

@ computing resduals @
ytilda = y - x*bpcce;                          @dep variable@
ff = ones(rows(ff),1)~ff;

ei = dyf - dxf*bpcce;
em = reshape(ei,nn,t);
ei = em';

{t_res,n_res,cdx} = CDP(ei);
"CD test";cdx;

/*********** Input Data *******************/
x_res = ei ; /*  T by N */
a_size=0.1;
x_resst=standx(x_res);
{t_res,n_res,cd_res}=CD(x_res);

/********** Exponent of cross-sectional dependence ************/
{rej,a_res,v_f_res,rej,rej,ggg_res,s_res,rej}=atildeall(x_resst,a_size);
om_res=1.65*((1/t_res)*(v_f_res)+(4/n_res)*(n_res^(1-a_res)*s_res/(ggg_res-1)))^(1/2)/(2*ln(n_res)); /* 90% CI: 1.65, 95% CI: 1.96 */


"Exponents of CSD (alpha) and 90% c.i. ";
"a_res"; (a_res-om_res)~a_res~(a_res+om_res);
"";"";

@ excel output for CSD test@
ret = xlsWritem(cdx, "2DPCCE", "c10", 1, 0); 
ret = xlsWritem((a_res-om_res)~a_res~(a_res+om_res), "2DPCCE", "c12", 1, 0); 


@ ***************** CTFE ************** @  

dyfm = y - myij - myjt - myit + myi + myj + myt - gmy;

dxfm = zeros(rows(y),k);
j=1; do while j<=k;
    dxfm[.,j] = x[.,j]  - mxit[.,j] - mxjt[.,j] - mxij[.,j]  + mxi[.,j] + mxj[.,j] + mxt[.,j] - gmx[.,j];
j=j+1;endo;

dxfm =  dxfm[.,3:k];

{bpcce, sehet, tstat} = FE1(dyfm,dxfm,nn,T,k);

bccep = bpcce~sehet~tstat;
PROUT = Xname1~bccep ;
" CTFE ";"             Coeff     se      t-ratio";
call printfm(PRout,0~1~1~1,"s"~8~8|"lf"~10~4|"lf"~10~4|"lf"~10~4); 
"";

@ output in excel   @
ret = xlsWritem(PROUT, "CTFE", "b2", 1, 0); 


ei = dyfm - dxfm*bpcce;      
em = reshape(ei,nn,t);
ei = em';

{t_res,n_res,cdx} = CDP(ei);
"CD test";cdx;

/*********** Input Data *******************/
x_res = ei ; /*  T by N */
a_size=0.1;
x_resst=standx(x_res);

/********** Exponent of cross-sectional dependence ************/
{rej,a_res,v_f_res,rej,rej,ggg_res,s_res,rej}=atildeall(x_resst,a_size);
om_res=1.65*((1/t_res)*(v_f_res)+(4/n_res)*(n_res^(1-a_res)*s_res/(ggg_res-1)))^(1/2)/(2*ln(n_res)); /* 90% CI: 1.65, 95% CI: 1.96 */

"";"";
" Exponents of CSD (alpha) and 90% c.i ";
"a_res"; (a_res-om_res)~a_res~(a_res+om_res);
"";"";

@ excel output for CSD test@
ret = xlsWritem(cdx, "CTFE", "c10", 1, 0); 
ret = xlsWritem((a_res-om_res)~a_res~(a_res+om_res), "CTFE", "c12", 1, 0); 

@****************** 3D PCCE ****************************@

@ factors selected for the  final version@
ff = mxt[.,3:4]~data[.,15]; @ 05 gen@


dyfm2 = dyfm; 
dxfm2 = dxfm;


ff2 = zeros(T,cols(ff));
j=1; do while j<=cols(ff);
    ff2[.,j] = ff[1:T,j] -  meanc(ff[1:T,j]).*.ones(T,1);
j=j+1;endo;

ff = ones(nn,1).*.ff2[1:T,.];

dyf  = trpnc(dyfm2); 

kk = cols(dxfm2);

dxf = zeros(rows(x),kk);
j=1; do while j<=kk;
  dxf[.,j] = trpnc(dxfm2[.,j]);
j=j+1;endo;


{bpcce, sehet, tstat} = FE1(dyf,dxf,nn,T,k);

bccep = bpcce~sehet~tstat;
PROUT = Xname1~bccep ;
" 3D PCCE   ";"             Coeff     se      t-ratio";
call printfm(PRout,0~1~1~1,"s"~8~8|"lf"~10~4|"lf"~10~4|"lf"~10~4); 
"";

@ output in excel   @
ret = xlsWritem(PROUT, "3DPCCE", "b2", 1, 0); 

@ residuals @
ei =   dyf - dxf*bpcce;   
em = reshape(ei,nn,t);
ei = em';

{t_res,n_res,cdx} = CDP2(ei);
"CD test";cdx;

x_res = ei ; /*  T by N */
a_size=0.1;
x_resst=standx(x_res);

/********** Exponent of cross-sectional dependence ************/
{rej,a_res,v_f_res,rej,rej,ggg_res,s_res,rej}=atildeall(x_resst,a_size);
om_res=1.65*((1/t_res)*(v_f_res)+(4/n_res)*(n_res^(1-a_res)*s_res/(ggg_res-1)))^(1/2)/(2*ln(n_res)); /* 90% CI: 1.65, 95% CI: 1.96 */



"";"";
"Exponents of CSD (alpha) and 90% c.i. ";
"a_res"; (a_res-om_res)~a_res~(a_res+om_res);
"";"";

@ excel output for CSD test@
ret = xlsWritem(cdx, "3DPCCE", "c10", 1, 0); 
ret = xlsWritem((a_res-om_res)~a_res~(a_res+om_res), "3DPCCE", "c12", 1, 0); 



@*******************************************************@
@************************ PROCS ************************@
@*******************************************************@


@************************   y_bar_.jt   ************************@

proc(1) = TRDJTS(y, nn, t);
local ini, fini, myjt, qy, tmpy, i, ysort, tmp,tmpsort;

ysort = sortc(y,2);
ini = 1;
fini = 0;

i = 1; 
do until  i > nn;
fini = fini + t;

tmpy  = ysort[ini:fini,1]~meanc(ysort[ini:fini,3]).*ones(t,1);

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + t;
i = i + 1;
endo;
  
tmp = qy;
tmpsort = sortc(tmp,1);

myjt = tmpsort[.,2];
retp(myjt);
endp;


@*********************   y_bar_i.t   ******************@
proc(1) = TRDITS(y, nn, t);
local ini, fini, myit, qy, tmpy, i, ysort, tmp,tmpsort;

ysort = sortc(y,2);
ini = 1;
fini = 0;

i = 1; 
do until  i > nn;
fini = fini + t;

tmpy  = ysort[ini:fini,1]~meanc(ysort[ini:fini,3]).*ones(t,1);

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + t;
i = i + 1;
endo;
  
tmp = qy;
tmpsort = sortc(tmp,1);

myit = tmpsort[.,2];
retp(myit);
endp;


@*** meanc of each time series =   y_bar_..t   ***@

proc(1) = TRDT(y, nn, t);
local ini, fini, myt, qy, tmpy, i, ysort, tmp,tmpsort;

ysort = sortc(y,4);

ini = 1;
fini = 0;

i = 1; 
do until  i > nn;
fini = fini + t;

tmpy  = ysort[ini:fini,1]~meanc(ysort[ini:fini,5]).*ones(t,1)~ysort[ini:fini,4];

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + t;
i = i + 1;
endo;
  
tmp = qy;

tmpsort = sortc(tmp,1);
myt = tmpsort[.,2];
retp(myt);
endp;


@*** meanc of each time series =   y_bar_.j.  SORT OK ***@

proc(1) = TRDJ(y, nn, t);
local ini, fini, myij, qy, tmpy, i, ysort, tmp,tmpsort;

ysort = sortc(y,3);

ini = 1;
fini = 0;
i = 1; 
do until  i > nn;
fini = fini + t;

tmpy  = ysort[ini:fini,1]~meanc(ysort[ini:fini,5]).*ones(t,1);

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + t;
i = i + 1;
endo;
  
tmp = qy;
tmpsort = sortc(tmp,1);

myij = tmpsort[.,2];
retp(myij);
endp;


@**** meanc of each i should be  y_bar_i.. OK  ***@

proc(1) = TRDJT(y, nn, t);

local ini, fini, myjt, qy, tmpy, i;


ini = 1;
fini = 0;
i = 1; 
do until  i > n;
fini = fini + ((n-1)*t);

tmpy  = meanc(y[ini:fini]).*ones(((n-1)*t),1);

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + ((n-1)*t);
i = i + 1;
endo;
  
myjt = qy;
retp(myjt);
endp;


@*** meanc of each time series =   y_bar_ij.   ***@

proc(1) = TRDIJ(y, nn, t);
local ini, fini, myij, qy, tmpy, i, ysort, tmp,tmpsort;

ysort = sortc(y,2);

ini = 1;
fini = 0;
i = 1; 
do until  i > nn;
fini = fini + t;

tmpy  = ysort[ini:fini,1]~meanc(ysort[ini:fini,3]).*ones(t,1);

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + t;
i = i + 1;
endo;
  
tmp = ind3~ind2ij~qy;
tmpsort = sortc(tmp,1);

myij = tmpsort[.,3];
retp(myij);
endp;

@*** meanc of each time series =   y_bar_ij.   ***@

proc(1) = TRDIJ(y, nn, t);

local ini, fini, myij, qy, tmpy, i;

ini = 1;
fini = 0;
i = 1; 
do until  i > nn;
fini = fini + t;

tmpy  = meanc(y[ini:fini,3]).*ones(t,1);

if i eq 1;
     qy  = tmpy;
   else;
     qy  = qy|tmpy;
   endif;

ini = ini + t;
i = i + 1;
endo;
  
myij = qy;
retp(myij);
endp;


@ ***************************** proc to demean as in FE ***************************** *@

proc trf(y);
local yfmp,yfp, ini, fini,y0,  inpt, xreg, tmpy,qy, tis, i;

ini  = 1;  @ initial position for ith group @
fini = 0;  @ end position for ith group @
i = 1;
do until i > nn;

 fini = fini + t;

  y0 = y[ini:fini];       							 @ y for ith group @
  inpt = ones(rows(y0),1); 				 		@ T by 1 vector of constants @

   xreg = inpt;    
   tmpy  = y0 - xreg*invpd(xreg'xreg)*xreg'y0; 		 @ demeaned, defactored y @
 
   if i eq 1;
     QY  = tmpy;
   else;
     QY  = QY|tmpy;
   endif;

ini = ini + t;
   i = i + 1;
endo;

 yfp = reshape(Qy,nn,t);
 yfmp = yfp[.,1:t];
retp(vec(yfmp'));

endp;


@ ***************************** proc to demean data Pesaran (2006) plus FE ***************************** *@

proc trp(y);

local yfmp,yfp, ini, fini,y0,  inpt, xreg, tmpy,qy, tis, i;
ini  = 1;  @ initial position for ith group @
fini = 0;  @ end position for ith group @
i = 1;
do until i > nn;


 fini = fini + t;

  y0 = y[ini:fini];       							 @ y for ith group @
  inpt = ones(rows(y0),1); 				 		@ T by 1 vector of constants @
  xreg = inpt~ff[1:rows(y0),.];    
      						   
   tmpy  = y0 - xreg*invpd(xreg'xreg)*xreg'y0; 		 @ demeaned, defactored y @
 
 if i eq 1;
     QY  = tmpy;
   else;
     QY  = QY|tmpy;
   endif;

ini = ini + t;
   i = i + 1;
endo;

 yfp = reshape(Qy,nj,nit);
 yfmp = yfp[.,1:nit];
retp(vec(yfmp'));

endp;


@ ***************************** proc to demean data Pesaran (2006) without FE ***************************** *@

proc trpnc(y);

local yfmp,yfp, ini, fini,y0,  inpt, xreg, tmpy,qy, tis, i;
ini  = 1;  @ initial position for ith group @
fini = 0;  @ end position for ith group @
i = 1;
do until i > nn;


 fini = fini + t;

  y0 = y[ini:fini];       							 @ y for ith group @
  inpt = ones(rows(y0),1); 				 		@ T by 1 vector of constants @
  xreg = ff[1:rows(y0),.];    
      						   
   tmpy  = y0 - xreg*invpd(xreg'xreg)*xreg'y0; 		 @ demeaned, defactored y @
 
 if i eq 1;
     QY  = tmpy;
   else;
     QY  = QY|tmpy;
   endif;

ini = ini + t;
   i = i + 1;
endo;

 yfp = reshape(Qy,nj,nit);
 yfmp = yfp[.,1:nit];
retp(vec(yfmp'));

endp;




@ ***************************** proc to two step demeaned data: Pesaran (2006) plus FE ***************************** *@

proc trpy(y);

local yfmp,yfp, ini, fini,y0,  inpt, xreg, tmpy,qy, tis, i, yi;
ini  = 1;  @ initial position for ith group @
fini = 0;  @ end position for ith group @
i = 1;
do until i > nn;


 fini = fini + t;

  y0 = y[ini:fini];       							 @ y for ith group @
  inpt = ones(rows(y0),1); 				 		@ T by 1 vector of constants @
  xreg = ff[1:rows(y0),.];    
  yi =  y[ini:fini] - myi[ini:fini];
      						   
   tmpy  = y0  + xreg*invpd(xreg'xreg)*xreg'yi; 		 @ demeaned, defactored y @
 
 if i eq 1;
     QY  = tmpy;
   else;
     QY  = QY|tmpy;
   endif;

ini = ini + t;
   i = i + 1;
endo;

 yfp = reshape(Qy,nj,nit);
 yfmp = yfp[.,1:nit];
retp(vec(yfmp'));

endp;


@ CD test *@
@ CD test from BKP *@
proc(3) = CDP(x);
local n,t,interc,coef,x_prel,m_x,std_x,little_x,rho_var_bar,cd;
t=rows(x);
n=cols(x);
interc=ones(t,1);
coef=inv(interc'*interc)*interc'*x;
x_prel=x-interc*coef; 
m_x=meanc(x_prel)';
std_x=stdc(x_prel)';
little_x=zeros(t,n); 
for j(1,n,1);
little_x[.,j]=x_prel[.,j]/std_x[1,j];
endfor;
rho_var_bar=(meanc(sumc(little_x').^2)-n)/(n*(n-1));
cd=sqrt((t)/2)*sqrt(n*(n-1))*rho_var_bar;
retp(t,n,cd);
endp;

@ CD test from BKP but adjusted to our case ??? *@
proc(3) = CDP3(x);
local n,t,interc,coef,x_prel,m_x,std_x,little_x,rho_var_bar,cd;
t=rows(x);
n=cols(x);
interc=ones(t,1);
coef=inv(interc'*interc)*interc'*x;
x_prel=x-interc*coef; 
m_x=meanc(x_prel)';
std_x=stdc(x_prel)';
little_x=zeros(t,n); 
for j(1,n,1);
little_x[.,j]=x_prel[.,j]/std_x[1,j];
endfor;
rho_var_bar=(sumc(sumc(little_x').^2)-n)/(n*(n-1));
cd=sqrt(2*t./n*(n-1))*rho_var_bar;
retp(t,n,cd);
endp;

@ CD TEST PESARAN MINE and STATA@
proc(3) = CDP1(x);
local n,t,cx, ro , cdro ,cd;
t=rows(x);
n=cols(x);
cx = corrx(ei);
ro =  lowmat(cx);
cdro = vec(ro);
cd = ((2*T)/(n*(n-1)))^0.5 * (sumc(cdro)-n);
retp(t,n,cd);
endp;

@ CD TEST ADJSTED  @
proc(3) = CDP2(x);
local n,t,cx, ro , cdro ,cd, cd1;
t=rows(x);
n=cols(x);
cx = corrx(ei);
ro =  lowmat(cx);
cdro = vec(ro);
cd = sqrt(2*T/(n*(n-1)))*(sumc(cdro)-n);
retp(t,n,cd);
endp;

/**********************************************/
proc(4) = FE(Y,X,N,T,k);
/**********************************************/

local bfep, Qres, Qobs, Qsig2, Qtemp, Qvar, Qse, Qt;

bfep = invpd(x'x)*x'y;

Qres  = y - x*bfep;          @ residual @
Qobs  = rows(y) - (n) - rows(bfep);                @ observation used @
Qsig2 = Qres'Qres/(Qobs);   @ within variance Qsig2 = Qres'Qres/(Qobs-nreg)@
Qtemp = invpd(x'x);                   
Qvar  = Qsig2.*Qtemp;
Qse   = sqrt(diag(Qvar));        @ standard error @
Qt    = bfep./Qse;             @ t-ratios @

retp(bfep,Qse,Qt, Qsig2); endp;


/**********************************************/
proc(4) = FE2w(Y,X,N,T,k);
/**********************************************/

local bfep, Qres, Qobs, Qsig2, Qtemp, Qvar, Qse, Qt;

bfep = invpd(x'x)*x'y;

Qres  = y - x*bfep;          @ residual @
Qobs  = rows(y) - (n-1) - (t-1) - rows(bfep);                @ observation used @
Qsig2 = Qres'Qres/(Qobs);   @ within variance Qsig2 = Qres'Qres/(Qobs-nreg)@
Qtemp = invpd(x'x);                   
Qvar  = Qsig2.*Qtemp;
Qse   = sqrt(diag(Qvar));        @ standard error @
Qt    = bfep./Qse;             @ t-ratios @

retp(bfep,Qse,Qt, Qsig2); endp;

@****************************@
proc(3) = FE1(Y,X,N,T,k);
@****************************@

local bfep, Qres, Qobs, Qsig2, Qtemp, Qvar, Qse, Qt, xxi, e;

bpcce = invpd(x'x)*x'y;

xxi = invpd(moment(x,0));
e = y - x*bpcce;
sehet = sqrt(diag(xxi*moment(x.*e,0)*xxi));

tstat = bpcce./sehet;
retp(bpcce, sehet, tstat); endp;




/* ******************************************************** Procedures BKS ******************************************************* */
/* Standardise data */
proc standx(x);
local n,t,m_x,std_x,x_stand;
t=rows(x);
n=cols(x);
m_x=meanc(x)';
std_x=stdc(x)';
x_stand=zeros(t,n);
for i(1,n,1);
x_stand[.,i]=(x[.,i]-m_x[1,i])/std_x[1,i];
endfor;
retp(x_stand);
endp;

/* Pesaran, M.H. (2015). Testing Weak Cross-sectional Dependence in Large Panels. Econometrics Reviews (forthcoming) */
/* CD test */
proc (3)=CD(x);
local n,t,interc,coef,x_prel,m_x,std_x,little_x,rho_var_bar,cd;
t=rows(x);
n=cols(x);
interc=ones(t,1);
coef=inv(interc'*interc)*interc'*x;
x_prel=x-interc*coef; 
m_x=meanc(x_prel)';
std_x=stdc(x_prel)';
little_x=zeros(t,n); 
for j(1,n,1);
little_x[.,j]=x_prel[.,j]/std_x[1,j];
endfor;
rho_var_bar=(meanc(sumc(little_x').^2)-n)/(n*(n-1));
cd=sqrt((t)/2)*sqrt(n*(n-1))*rho_var_bar;
retp(t,n,cd);
endp;



end;
